import sys
from setuptools import setup

setup(
    name='wavelets_pytorch',
    version='0.1',
    author='Tom Runia',
    author_email='tomrunia@gmail.com',
    url='https://github.com/tomrunia/PyTorchWavelets',
    description='Wavelet Transform in PyTorch',
    long_description='Fast CPU/CUDA implementation of the Continuous Wavelet Transform in PyTorch.',
    license='MIT',
    packages=['wavelets_pytorch'],
    scripts=[]
)